char *s = N_("%%");
char *s = N_("%A");
char *s = N_("%B");
char *s = N_("%H");
char *s = N_("%I");
char *s = N_("%M");
char *s = N_("%S");
char *s = N_("%U");
char *s = N_("%W");
char *s = N_("%X");
char *s = N_("%Y");
char *s = N_("%Z");
char *s = N_("%a");
char *s = N_("%b");
char *s = N_("%c");
char *s = N_("%d");
char *s = N_("%j");
char *s = N_("%m");
char *s = N_("%p");
char *s = N_("%w");
char *s = N_("%x");
char *s = N_("%y");
char *s = N_("50");
char *s = N_("<b>All Available icons</b>");
char *s = N_("<b>Date &amp; time formatting key</b>");
char *s = N_("<b>Date and Time</b>");
char *s = N_("<b>General KeepNote Options</b>");
char *s = N_("<b>Manage icons</b>");
char *s = N_("<b>Node icons</b>");
char *s = N_("<b>Notebook Update</b>");
char *s = N_("<b>Quick pick icons</b>");
char *s = N_("<b>Startup</b>");
char *s = N_("<b>This Notebook</b>");
char *s = N_("<i>note: backup and upgrade may take a while for larger notebooks.</i>");
char *s = N_("A literal \"%\" character");
char *s = N_("Alternative index location:");
char *s = N_("Autosave:");
char *s = N_("Browse...");
char *s = N_("Case _Sensitive");
char *s = N_("Day of the month as a decimal number [01,31]");
char *s = N_("Day of the year as a decimal number [001,366]");
char *s = N_("Default Notebook:");
char *s = N_("Default font:");
char *s = N_("Different year:");
char *s = N_("Find Text:");
char *s = N_("Find/Replace");
char *s = N_("From here, you can customize the format of timestamps in the listview for 4 different scenarios (e.g. how to display a timestamp from today, this month, or this year)");
char *s = N_("Hide from taskbar");
char *s = N_("Hour (12-hour clock) as a decimal number [01,12]");
char *s = N_("Hour (24-hour clock) as a decimal number [00,23]");
char *s = N_("Keep aspect ratio");
char *s = N_("KeepNote Preferences");
char *s = N_("Locale's abbreviated month name");
char *s = N_("Locale's abbreviated weekday name");
char *s = N_("Locale's appropriate date and time representation");
char *s = N_("Locale's appropriate date representation");
char *s = N_("Locale's appropriate time representation");
char *s = N_("Locale's equivalent of either AM or PM");
char *s = N_("Locale's full month name");
char *s = N_("Locale's full weekday name");
char *s = N_("Minute as a decimal number [00,59]");
char *s = N_("Month as a decimal number [01,12]");
char *s = N_("New Icon");
char *s = N_("No Default Notebook");
char *s = N_("Notebook Update");
char *s = N_("Notebook-specific Icons");
char *s = N_("Open Last Notebook");
char *s = N_("Replace Text:");
char *s = N_("Replace _All");
char *s = N_("Resize Image");
char *s = N_("Same day:");
char *s = N_("Same month:");
char *s = N_("Same year:");
char *s = N_("Sea_rch Forward");
char *s = N_("Search _Backward ");
char *s = N_("Second as a decimal number [00,61]");
char *s = N_("Snap:");
char *s = N_("Standard Icons");
char *s = N_("This notebook has format version 1 and must be updated to version 2 before openning.");
char *s = N_("Time zone name (no characters if no time zone exists)");
char *s = N_("Use current notebook");
char *s = N_("Use sysem tray icon");
char *s = N_("Week number of the year (Monday as the first day of the week) \n"
             "as a decimal number [00,53]. All days in a new year preceding \n"
             "the first Monday are considered to be in week 0");
char *s = N_("Week number of the year (Sunday as the first day of the week) \n"
             "as a decimal number [00,53]. All days in a new year preceding \n"
             "the first Sunday are considered to be in week 0");
char *s = N_("Weekday as a decimal number [0(Sunday),6]");
char *s = N_("Year with century as a decimal number");
char *s = N_("Year without century as a decimal number [00,99]");
char *s = N_("_Apply");
char *s = N_("_Cancel");
char *s = N_("_Close");
char *s = N_("_Find");
char *s = N_("_Ok");
char *s = N_("_Replace");
char *s = N_("delete icon");
char *s = N_("gtk-cancel");
char *s = N_("gtk-ok");
char *s = N_("gtk-revert-to-saved");
char *s = N_("height:");
char *s = N_("icon:");
char *s = N_("open-icon:");
char *s = N_("save backup before updating (recommended)");
char *s = N_("seconds");
char *s = N_("set icon");
char *s = N_("set open-icon");
char *s = N_("width:");
